#!/usr/bin/perl
#
# File:  "$Id: patch_installer.pl,v 1.31 2013/09/24 09:46:54 berit_wang Exp $"
#
# Description:
#
# BUGS:
#
# CHANGES:
#
# NOTE:
#
# First Created on  2010/03/05 by jonas_liu@openfind.com.tw
# Last Modified on "$Date: 2013/09/24 09:46:54 $" by  $Author: berit_wang $
#
# Copyright (c) Openfind Information Technology INC.
# All rights reserved.
#
use strict;
use Getopt::Std;
use Digest::MD5 qw(md5_base64);
use File::Basename;
use POSIX;

my $CVSREV = '$Revision: 1.31 $';
my $CVSID = substr($CVSREV, 11, -2);

$ENV{PATH} = "/webmail/tools:/usr/local/bin:/bin:/usr/bin:/usr/sbin";
$| = 1;

my $VERSION = join("\n",
	"",
	"Mail2000 V60 Patch Installer v$CVSID",
	"Copyright (C) CyberSolutions Inc.",
	"All rights reserved.\n",
	""
);
my $DEBUG_MODE = 0;
my $QUIET_MODE = 0;

my $DEF_OWNER = "webmail";
my $DEF_GROUP = "webmail";
my $RUNNER = "webmail";
my $OWNER = $DEF_OWNER;
my $GROUP = $DEF_GROUP;
my $RUNNER_UID = 0;
my $OWNER_UID = 0;
my $GROUP_GID = 0;

my $M2K_PATCH_PROFILE = "patch.info";
my $M2K_PATCH_INFO = "/webmail/etc/m2kpatch.info";
my $M2K_PATCH_CONF = "m2kpatch.conf";
my $M2K_BACKUP_BASEDIR = "/webmail/backup";
my $POST_INST_SCRIPT_DIR = "/webmail/bin";

my $M2K_WEBMAIL_HOME = "/webmail";
my $M2K_TMP_DIR = "/webmail/tmp";
my $M2K_DAEMON_TOOL = "/webmail/tools/m2kctrl";
my $M2K_PATCH_TOOL = "/webmail/tools/update_info_mgr";
my $M2K_SHM_TOOL = "/webmail/tools/restartshm";
my $M2K_VERSION_TOOL = "/webmail/tools/version_info";

my $CP_CMD = "cp -pf";
my $MKDIR_CMD = "mkdir";
my $LN_CMD = "ln";
my $CHOWN_CMD = "chown";
my $CHMOD_CMD = "chmod";
my $MV_CMD = "mv -f";
my $DF_CMD = "df -kP";
my $DF_CMD_SOLARIS = "df -k";
my $RM_CMD = "rm";
my $TAR_CMD = "tar";
my $GZIP_CMD = "gzip";

my ($_SEC, $_MIN, $_HOUR, $_MDAY, $_MON, $_YEAR) = localtime(time);
my %g_Opt;
my @g_ChangeLog;
my %g_SpaceNeeded;

my $PATCH_BASEDIR = getcwd();
my $PATCH_ROLLBACK_TMP = "rollback_tmp.$$";
my $PATCH_NAME = undef;
my $PATCH_DESC = undef;
my $PATCH_STATUS_FILE = undef;
my $PATCH_POST_INST_MODE = 0;
my $PATCH_ROOT_PERM = 0;

use constant FIN_R_NONE => 0;
use constant FIN_R_SUCC => 1;
use constant FIN_R_ERROR => 2;

use constant FIN_S_ERR_NONE => 0;
use constant FIN_S_ERR_DAEMON_CHK => 1;
use constant FIN_S_ERR_PATCH_DEPCHK => 2;
use constant FIN_S_ERR_INTEGRITY_CHK => 3;
use constant FIN_S_ERR_STORAGE_CHK => 4;
use constant FIN_S_ERR_BACKUP => 5;
use constant FIN_S_ERR_INTERNAL => 6;
use constant FIN_S_ERR_COPY_FILE => 7;
use constant FIN_S_ERR_MOVE_FILE => 8;
use constant FIN_S_ERR_REMOVE_FILE => 9;
use constant FIN_S_ERR_CREATE_DIR => 10;
use constant FIN_S_ERR_CREATE_LINK => 11;
use constant FIN_S_ERR_EXECMD => 12;
use constant FIN_S_ERR_PATCH_POST => 13;
use constant FIN_S_ERR_SET_PERM => 14;
use constant FIN_S_ERR_FILE_DEPCHK => 15;
use constant FIN_S_ERR_SYSCHK => 16;
use constant FIN_S_ERR_USER_ABORT => 50;
use constant FIN_S_ERR_UNKNOWN => 100;

sub _DEBUG {
	return if(!$DEBUG_MODE);
	print STDERR "[DEBUG] ", @_;
}

sub _MESG {
	return if($QUIET_MODE);
	print STDOUT @_;
}

sub _WARN {
	print STDOUT "[WARN] ", @_;
}

sub _ERR {
	print STDERR "\n===[ ERROR ]===\n\n", @_, "\n";
}

sub _FINISH {
	my ($status, $status_code, $mesg) = @_;

	if($DEBUG_MODE) {
		_MESG("\n");
		_MESG("=====[ START OF CHANGE LOG ]=====\n");
		s_DumpChangeLog();
		_MESG("=====[ END OF CHANGE LOG ]=====\n\n");
	}

	if(!defined($PATCH_STATUS_FILE)) {	
		if($status == FIN_R_ERROR) {
			_ERR("code=$status_code\n$mesg\n");
		} else {
			_MESG("$mesg\n");
		}
	}
	else {
		if(!open(FP, "> $PATCH_STATUS_FILE")) {
			_WARN("cannot open status file: $PATCH_STATUS_FILE: $!\n");
			$PATCH_STATUS_FILE = undef;
			_FINISH($status, $status_code, $mesg);
		}
		else {
			printf FP "Status = %d\n", ($status == FIN_R_SUCC) ? 1 : 2;
			printf FP "FailStatus = %d\n", $status_code;
			printf FP "Message = %s\n", $mesg;
			close(FP);
		}
	}
	
	exit($status == FIN_R_SUCC ? 0 : 1);
}

sub s_CheckIgnoreError {
	my ($ignore_err, $status_code, $mesg) = @_;
	
	if(0 == $ignore_err) {
		_FINISH(FIN_R_ERROR, $status_code, $mesg);
	}
	else {
		_WARN(sprintf("%s, code=%d\n", $mesg, $status_code));
	}
	
	return undef;
}

sub s_GetOSType {
   my $szUname = `uname -a`;
        
   return "solaris"	if($szUname =~ m/SunOS/i);
   return "freebsd"	if($szUname =~ m/FreeBSD/i);
   return "linux"		if($szUname =~ m/Linux/i);
   return "Unknown";
}

sub s_RunCmd {
	return system(@_);
}

sub s_CheckFilePath {
	my $path = shift;
	
	return 0 if($path =~ /[\'\"\<\>\|\$]/);
	return 1;
}

sub s_GetStorageFreeSpace {
	my ($func_code, $path) = @_;
	my $ret;

	while (! -d $path) {
		$path = dirname($path);
		_FINISH(FIN_R_ERROR, $func_code, "the direocty to check storage free space is not found.")
			if (!$path);
	}

	if (s_GetOSType() eq "solaris") {
	   $ret = `$DF_CMD_SOLARIS \"$path\" | tail -1`;
	} else {
	   $ret = `$DF_CMD \"$path\" | tail -n 1`;
	}
	
	_FINISH(FIN_R_ERROR, $func_code, "execute df failed.")
		if(0 != WEXITSTATUS($?));

	_FINISH(FIN_R_ERROR, $func_code, "unknown result of df: $ret")
		if($ret !~ /^(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)$/);

	# return (mount_point, free_space_in_kb)
	return ($6, int($4));
}

sub s_GetPostInstScriptPath {
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "unknown patch name when decide post-inst script path.")
		if(!defined($PATCH_NAME));
		
	return sprintf("$POST_INST_SCRIPT_DIR/post_$PATCH_NAME.pl");
}

sub s_AddChangeLog {
	my ($type, $param) = @_;
	
	_DEBUG("CHANGE-LOG: $type\n");
	
	push(@g_ChangeLog, [$type, $param]);
}

sub s_AddPostScript {
   my @cmds = @_;
   
   my $script_file = s_GetPostInstScriptPath();
   
   if(!open(FP, ">> $script_file")) {
       _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "cannot open post-inst script file: $!");
   }
   print FP @cmds, "\n";
   close(FP);
   
   _DEBUG("[POST_SCRIPT] ", @cmds, "\n");
}

sub s_FileMD5 {
	my $path = shift;
	my $file_md5;
	
	open(FPMD5, $path) || return undef;
	binmode(FPMD5);
	$file_md5 = Digest::MD5->new->addfile(*FPMD5)->hexdigest;
	close(FPMD5);
	
	return $file_md5;
}

sub s_DumpChangeLog {
	my @log_text;

	foreach my $info (@g_ChangeLog) {
		my $type = $info->[0];
		my $param = $info->[1];
	
		if($type eq "CREATE_DIR") {
			push(@log_text, sprintf("create dir: %s, perm=%s", $param->{'path'}, $param->{'perm'}));
		}
		if($type eq "CREATE_DIR_H") {
			push(@log_text, sprintf("create dir: %s, perm=%s", $param->{'path'}, $param->{'perm'}, $param->{'owner'}, $param->{'group'}));
		}
		elsif($type eq "CREATE_LINK") {
			push(@log_text, sprintf("create link: %s => %s", $param->{'source'}, $param->{'target'}));
		}
		elsif($type eq "SET_PERM") {
			push(@log_text, sprintf("set perm: %s, perm=%s", $param->{'path'}, $param->{'perm'}));
		}
		elsif($type eq "SET_OWNERGROUP") {
			push(@log_text, sprintf("set owner-group: %s, owner=%s, group=%s", $param->{'path'}, $param->{'owner'}, $param->{'group'}));
		}
		elsif($type eq "COPY_FILE") {
			push(@log_text, sprintf("copy file: %s => %s", $param->{'source'}, $param->{'target'}));
		}
		elsif($type eq "MOVE_FILE") {
			push(@log_text, sprintf("move file: %s => %s", $param->{'source'}, $param->{'target'}));
		}
		elsif($type eq "REMOVE_FILE") {
			push(@log_text, sprintf("remove file: %s, rollback=%s", $param->{'path'}, $param->{'rollback'}));
		}
		elsif($type eq "EXEC_CMD") {
			push(@log_text, sprintf("exec: %s", $param->{'cmd'}));
		}
		else {
			push(@log_text, "unknown change log: $type");
		}
	}
	print join("\n", @log_text), "\n";
}

sub s_SetProperty {
	my ($ignore_err, $status_code, $operm, $ouid, $ogid, $dst_path) = @_;
	my $ret;

	$ret = chown($ouid, $ogid, $dst_path);
	if (0 == $ret) {
		my $ouname = getpwuid($ouid);
		my $ogname = getgrgid($ogid);

		return s_CheckIgnoreError($ignore_err, $status_code, "cannot change file ownerships: $dst_path => $ouid:$ogid ($ouname:$ogname)");
	}

	$ret = chmod($operm, $dst_path);
	return s_CheckIgnoreError($ignore_err, $status_code, "cannot change file permissions: $dst_path => $operm")
		if(0 == $ret);

	return 0;
}

sub s_CreateDirRecur {
	my ($dir_name, $perm, $owner, $group) = @_;
	my $ret;
	my $cmd;
	my $pdir;

	if(-d $dir_name) {
		return 0;
	}

	$pdir = dirname($dir_name);
	if ($pdir ne '.' && $pdir ne '/') {
		$ret = s_CreateDirRecur($pdir, $perm, $owner, $group);
		if ($ret != 0) {
			return $ret;
		}
	}

	$cmd = "$MKDIR_CMD -m $perm \"$dir_name\"";
	$ret = s_RunCmd($cmd);
	if ($ret != 0) {
		return $ret;
	}

	$cmd = "$CHOWN_CMD $owner:$group \"$dir_name\"";
	$ret = s_RunCmd($cmd);
	if ($ret != 0) {
		return $ret;
	}

	return 0;
}

sub s_DoUnlink {
	my ($func_code, $ignore_err, $path) = @_;
	my $ret;
	
	return (0, undef) if(! -e $path);

	my $md5 = s_FileMD5($path);
	
	return (0, s_CheckIgnoreError($ignore_err, $func_code, "cannot calc MD5 of $path"))
		if(!defined($md5));

	# check rollback temp directory
	if(! -e $PATCH_ROLLBACK_TMP) {
		$ret = s_RunCmd("$MKDIR_CMD -m 0700 -p \"$PATCH_ROLLBACK_TMP\"");
	
		return (0, s_CheckIgnoreError($ignore_err, $func_code, "cannot create rollback temp directory: $PATCH_ROLLBACK_TMP"))
			if($ret != 0);

		$ret = s_RunCmd("$CHOWN_CMD $OWNER:$GROUP \"$PATCH_ROLLBACK_TMP\"");
	
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot change owner/group of rollback temp directory: $PATCH_ROLLBACK_TMP")
			if($ret != 0);
	}

	return (0, s_CheckIgnoreError($ignore_err, $func_code, "rollback tmp: $PATCH_ROLLBACK_TMP is not a directory."))
		if(! -d $PATCH_ROLLBACK_TMP);

	# backup the file into rollback temp	
	$ret = s_RunCmd("$CP_CMD \"$path\" \"$PATCH_ROLLBACK_TMP/$md5\"");
	 
	return (0, s_CheckIgnoreError($ignore_err, $func_code, "cannot make rollback backup: $path (md5=$md5), ret=$ret"))
		if(0 != $ret);
		
	# do unlink
	$ret = unlink($path);
	
	return ($ret, $md5);
}

sub Cmd_CheckDaemon {
	my $daemon_name = shift;
	my ($ret, $ret_status);
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_DAEMON_CHK, "cannot execute mail2000 daemon control tool.")
		if(! -x $M2K_DAEMON_TOOL);
	
	$ret = `$M2K_DAEMON_TOOL -c status -s $daemon_name`;

	_FINISH(FIN_R_ERROR, FIN_S_ERR_DAEMON_CHK, "unknown status from mail2000 daemon control tool.")
		if($ret !~ /is (stop|running|partial running)/);

	$ret_status = $1;

	return if($ret_status eq "stop");

	_FINISH(FIN_R_ERROR, FIN_S_ERR_DAEMON_CHK, "daemon $daemon_name is running.");
}

sub Cmd_CreateDir {
	my ($ignore_err, $perm, $dir_name) = @_;
	my $ret;
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_DIR, "invalid dir path: $dir_name")
		if(0 == s_CheckFilePath($dir_name));
		
	if(-d $dir_name) {
		_DEBUG("create dir: $dir_name already exists.\n");
		return;
	}
	
	my $cmd = "$MKDIR_CMD -m $perm -p \"$dir_name\"";

	# CreateDir command does not support post-install mode because the directory should be created at first for other commands.
	$ret = s_RunCmd($cmd);

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_DIR, "cannot create dir: $dir_name, ret=$ret")
		if(0 != $ret);

	s_AddChangeLog("CREATE_DIR", {
		'perm' => $perm,
		'path' => $dir_name
	});
}

sub Cmd_CreateDirHier {
	my ($ignore_err, $perm, $dir_name, $owner, $group) = @_;
	my $ret;

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_DIR, "invalid dir path: $dir_name")
		if(0 == s_CheckFilePath($dir_name));

	if(-d $dir_name) {
		_DEBUG("create dir: $dir_name already exists.\n");
		return;
	}

	if (!defined($owner)) {
		$owner = $DEF_OWNER;
	}
	if (!defined($group)) {
		$group = $DEF_GROUP;
	}

	# CreateDirHier command does not support post-install mode because the directory should be created at first for other commands.
	$ret = s_CreateDirRecur($dir_name, $perm, $owner, $group);

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_DIR, "cannot create dir hierarchically: $dir_name, ret=$ret")
		if(0 != $ret);

	s_AddChangeLog("CREATE_DIR_H", {
		'perm' => $perm,
		'path' => $dir_name,
		'owner' => $owner,
		'group' => $group
	});
}

sub Cmd_SetPerm {
	my ($ignore_err, $path, $owner, $group, $perm) = @_;
	my $ret;
	my $cmd;

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_SET_PERM, "invalid path: $path")
		if(0 == s_CheckFilePath($path));
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_SET_PERM, "path is not found: $path")
		if(!$PATCH_POST_INST_MODE && (! -e $path));

	if (!defined($owner)) {
		$owner = $DEF_OWNER;
	}
	if (!defined($group)) {
		$group = $DEF_GROUP;
	}

	# set owner/group
	$cmd = "$CHOWN_CMD $owner:$group \"$path\"";
	
	if($PATCH_POST_INST_MODE) {
		s_AddPostScript($cmd);
	}
	else {
		$ret = s_RunCmd($cmd);
		
		if(0 != $ret) {
			s_CheckIgnoreError($ignore_err, FIN_S_ERR_SET_PERM, "cannot change owner/group to $owner/$group: $path");
		}
		else {
			s_AddChangeLog("SET_OWNERGROUP", {
				'owner' => $owner,
				'group' => $group,
				'path' => $path
			});
		}
	}

	# set mode
	$cmd = "$CHMOD_CMD $perm \"$path\"";

	if($PATCH_POST_INST_MODE) {
		s_AddPostScript($cmd);
	}
	else {
		$ret = s_RunCmd($cmd);
	
		if(0 != $ret) {
			s_CheckIgnoreError($ignore_err, FIN_S_ERR_SET_PERM, "cannot change perm to $perm: $path")
		}
		else {
			s_AddChangeLog("SET_PERM", {
				'perm' => $perm,
				'path' => $path
			});
		}
	}
}

sub Cmd_CreateLink {
	my ($ignore_err, $keep_pty, $src_path, $dst_path) = @_;
	my ($ret, $md5);
	my $org_dst_path;
	my @stat;
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "invalid src-path: $src_path")
		if(0 == s_CheckFilePath($src_path));

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "invalid dst-path: $dst_path")
		if(0 == s_CheckFilePath($dst_path));
		
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "src-path is not found: $src_path")
		if(! -e $src_path);

	if(-e $dst_path) {
		# get original file properties
		if ($keep_pty == 1) {
			@stat = stat($dst_path);

			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "cannot get original file status: $dst_path")
				if (scalar(@stat) == 0);
		}

		_DEBUG("dst-path exists: $dst_path, try to remove it.\n");
		
		if($PATCH_POST_INST_MODE) {
			s_AddPostScript("$RM_CMD -f \"$dst_path\"");
		}
		else {
			($ret, $md5) = s_DoUnlink(FIN_S_ERR_CREATE_LINK, $ignore_err, $dst_path);
			_DEBUG("create-link, unlink-to-rollback-tmp: ret=$ret, md5=$md5\n");
			
			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "cannot remove exist file, when link $src_path to $dst_path.")
				if(-e $dst_path);

			s_AddChangeLog("REMOVE_FILE", {
				'path' => $dst_path,
				'rollback' => $md5
			});
		}
	}
	
	if($PATCH_POST_INST_MODE) {
		$org_dst_path = $dst_path;
		$dst_path = "$org_dst_path.post-inst";
	}
		
	$ret = s_RunCmd("$LN_CMD -s \"$src_path\" \"$dst_path\"");
	 
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_CREATE_LINK, "cannot create link: $src_path => $dst_path, ret=$ret")
		if(0 != $ret);
		
	# set original file properties to destination file
	if ($keep_pty == 1 && scalar(@stat) > 0) {
		$ret = s_SetProperty($ignore_err, FIN_S_ERR_CREATE_LINK, $stat[2] & 07777, $stat[4], $stat[5], $dst_path);
		return $ret if(0 != $ret);
	}

	s_AddChangeLog("CREATE_LINK", {
		'source' => $src_path,
		'target' => $dst_path
	});

	if($PATCH_POST_INST_MODE) {
		s_AddPostScript("$MV_CMD \"$dst_path\" \"$org_dst_path\"");
	}
}

sub Cmd_CopyFile {
	my ($ignore_err, $keep_pty, $src_path, $dst_path) = @_;
	my ($ret, $md5);
	my $org_dst_path;
	my @stat;
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "invalid src-path: $src_path")
		if(0 == s_CheckFilePath($src_path));

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "invalid dst-path: $dst_path")
		if(0 == s_CheckFilePath($dst_path));
		
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "src-path is not found: $src_path")
		if(! -e $src_path);

	if(-e $dst_path) {
		# get original file properties
		if ($keep_pty == 1) {
			@stat = stat($dst_path);

			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "cannot get original file status: $dst_path")
				if (scalar(@stat) == 0);
		}

		_DEBUG("dst-path exists: $dst_path, try to remove it.\n");

		if($PATCH_POST_INST_MODE) {
			s_AddPostScript("$RM_CMD -f \"$dst_path\"");
		}
		else {
			($ret, $md5) = s_DoUnlink(FIN_S_ERR_COPY_FILE, $ignore_err, $dst_path);
			_DEBUG("copy-file, unlink-to-rollback-tmp: ret=$ret, md5=$md5\n");
			
			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "cannot remove exist file, when copy $src_path to $dst_path.")
				if(-e $dst_path);

			s_AddChangeLog("REMOVE_FILE", {
				'path' => $dst_path,
				'rollback' => $md5
			});
		}
	}

	if($PATCH_POST_INST_MODE) {
		$org_dst_path = $dst_path;
		$dst_path = "$org_dst_path.post-inst";
	}

	$ret = s_RunCmd("$CP_CMD \"$src_path\" \"$dst_path\"");
	 
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_COPY_FILE, "cannot copy file: $src_path => $dst_path, ret=$ret")
		if(0 != $ret);

	# set original file properties to destination file
	if ($keep_pty == 1 && scalar(@stat) > 0) {
		$ret = s_SetProperty($ignore_err, FIN_S_ERR_COPY_FILE, $stat[2] & 07777, $stat[4], $stat[5], $dst_path);
		return $ret if(0 != $ret);
	}

	s_AddChangeLog("COPY_FILE", {
		'source' => $src_path,
		'target' => $dst_path
	});

	if($PATCH_POST_INST_MODE) {
		s_AddPostScript("$MV_CMD \"$dst_path\" \"$org_dst_path\"");
	}
}

sub Cmd_MoveFile {
	my ($ignore_err, $keep_pty, $src_path, $dst_path) = @_;
	my ($ret, $md5);
	my $org_dst_path;
	my @stat;
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "invalid src-path: $src_path")
		if(0 == s_CheckFilePath($src_path));

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "invalid dst-path: $dst_path")
		if(0 == s_CheckFilePath($dst_path));
		
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "src-path is not found: $src_path")
		if(! -e $src_path);

	if(-e $dst_path) {
		# get original file properties
		if ($keep_pty == 1) {
			@stat = stat($dst_path);

			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "cannot get original file status: $dst_path")
				if (scalar(@stat) == 0);
		}

		_DEBUG("dst-path exists: $dst_path, try to remove it.\n");

		if($PATCH_POST_INST_MODE) {
			s_AddPostScript("$RM_CMD -f \"$dst_path\"");
		}
		else {
			($ret, $md5) = s_DoUnlink(FIN_S_ERR_MOVE_FILE, $ignore_err, $dst_path);
			_DEBUG("move-file, unlink-to-rollback-tmp: ret=$ret, md5=$md5\n");
			
			return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "cannot remove exist file, when move $src_path to $dst_path.")
				if(-e $dst_path);

			s_AddChangeLog("REMOVE_FILE", {
				'path' => $dst_path,
				'rollback' => $md5
			});
		}
	}

	if($PATCH_POST_INST_MODE) {
		$org_dst_path = $dst_path;
		$dst_path = "$org_dst_path.post-inst";
	}
		
	$ret = s_RunCmd("$MV_CMD \"$src_path\" \"$dst_path\"");
	 
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_MOVE_FILE, "cannot move file: $src_path => $dst_path, ret=$ret")
		if(0 != $ret);

	# set original file properties to destination file
	if ($keep_pty == 1 && scalar(@stat) > 0) {
		$ret = s_SetProperty($ignore_err, FIN_S_ERR_MOVE_FILE, $stat[2] & 07777, $stat[4], $stat[5], $dst_path);
		return $ret if(0 != $ret);
	}

	s_AddChangeLog("MOVE_FILE", {
		'source' => $src_path,
		'target' => $dst_path
	});

	if($PATCH_POST_INST_MODE) {
		s_AddPostScript("$MV_CMD \"$dst_path\" \"$org_dst_path\"");
	}
}

sub Cmd_RemoveFile {
	my ($ignore_err, $path) = @_;
	my ($ret, $md5);
	
	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_REMOVE_FILE, "invalid path: $path")
		if(0 == s_CheckFilePath($path));

	return s_CheckIgnoreError($ignore_err, FIN_S_ERR_REMOVE_FILE, "src-path is not found: $path")
		if(!$PATCH_POST_INST_MODE && (! -e $path));
		
	if($PATCH_POST_INST_MODE) {
		s_AddPostScript("$RM_CMD -f \"$path\"");
	}
	else {
		($ret, $md5) = s_DoUnlink(FIN_S_ERR_REMOVE_FILE, $ignore_err, $path);
		_DEBUG("remove-file, rollback-tmp: ret=$ret, md5=$md5\n");
		 
		return s_CheckIgnoreError($ignore_err, FIN_S_ERR_REMOVE_FILE, "cannot remove file: $path: $!")
			if(1 != $ret);

		s_AddChangeLog("REMOVE_FILE", {
			'path' => $path,
			'rollback' => $md5
		});
	}
}

sub Cmd_ExeCmd {
	my ($ignore_err, $cmd) = @_;
	my $ret;

	if($PATCH_POST_INST_MODE) {
		s_AddPostScript($cmd);
	}
	else {
		$ret = s_RunCmd($cmd);

		if(0 != $ret) {
			s_CheckIgnoreError($ignore_err, FIN_S_ERR_EXECMD, "execute cmd failed: $cmd, ret=$ret")
		}
		else {
			s_AddChangeLog("EXEC_CMD", {
				'cmd' => $cmd
			});
		}
	}
}

sub Cmd_ExeCmdByInstaller {
	my ($ignore_err, $cmd) = @_;
	my $ret;

	$ret = s_RunCmd($cmd);

	if(0 != $ret) {
		s_CheckIgnoreError($ignore_err, FIN_S_ERR_EXECMD, "execute cmd by installer failed: $cmd, ret=$ret")
	}
	else {
		s_AddChangeLog("EXEC_CMD_I", {
			'cmd' => $cmd
		});
	}
}

sub Cmd_CheckPatchDep {
	my ($patch_name, $hint_text) = @_;
	my $ret;
	my $status;
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_DEPCHK, "cannot execute mail2000 patch info tool.")
		if(! -x $M2K_PATCH_TOOL);

	if ($RUNNER_UID == 0) {
		$ret = `su $OWNER -c '$M2K_PATCH_TOOL -q \"$patch_name\"'`;
	} else {
		$ret = `$M2K_PATCH_TOOL -q \"$patch_name\"`;
	}
	chomp $ret;

	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_DEPCHK, "unknown status from mail2000 patch info tool.")
		if($ret !~ /^(yes|no)/i);

	$status = $1;

	return if($status =~ /^yes$/i);
	
	_MESG("\n$hint_text\n\n")
		if ($hint_text ne "");

	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_DEPCHK, "patch dependency check failed: $patch_name, status=$status");
}

sub Cmd_CheckFileDep {
	my ($file_name, $version) = @_;
	my $ret_code;
	my $ret;
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_FILE_DEPCHK, "cannot execute mail2000 version info tool.")
		if(! -x $M2K_VERSION_TOOL);
	
	$ret = `$M2K_VERSION_TOOL \"$file_name\" $version`;

	_FINISH(FIN_R_ERROR, FIN_S_ERR_FILE_DEPCHK, "unknown status from mail2000 version info tool.")
		if($ret !~ /^(\d+)/);

	$ret_code = $1;

	return if($ret_code == 1);

	_FINISH(FIN_R_ERROR, FIN_S_ERR_FILE_DEPCHK, "file dependency check failed: $file_name ($version)")
		if($ret_code == 0);
}

sub Cmd_EvalIF {
	my $eval_cond = shift;

	my $func = eval("sub { return ($eval_cond) ? 1 : 0; }");
	return &$func();
}

sub Cmd_CheckStroageSpace {
	my ($path, $space_kbytes) = @_;
	my $ret;
	my ($mount_point, $free_space_needed_prev);
	my ($free_space, $free_space_needed);

	($mount_point, $free_space) = s_GetStorageFreeSpace(FIN_S_ERR_STORAGE_CHK, $path);
	$free_space_needed = int($space_kbytes) + 1024;
	$free_space_needed_prev = $g_SpaceNeeded{$mount_point};
	
	_DEBUG("check storage space of $path, mounted=$mount_point, space=$free_space, needed=$free_space_needed, prev_needed=$free_space_needed_prev\n");

	if($free_space >= $free_space_needed + $free_space_needed_prev) {
		$g_SpaceNeeded{$mount_point} += $free_space_needed;
		return;
	}

	_FINISH(FIN_R_ERROR, FIN_S_ERR_STORAGE_CHK, "storage space is not enough free=$free_space, needed=$free_space_needed, prev_needed=$free_space_needed_prev.")
}

sub Cmd_CheckFileMD5 {
	my ($path, $md5) = @_;
	my $file_md5;
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTEGRITY_CHK, "the file needed to check is not found: $path")
		if(! -e $path);
		
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTEGRITY_CHK, "the target needed to check is not a file: $path")
		if(! -f $path);
	
	$file_md5 = s_FileMD5($path);
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTEGRITY_CHK, "cannot calc MD5 of file $path")
		if(!defined($file_md5));

	return if($file_md5 eq $md5);
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTEGRITY_CHK, "integrity check failed: $path: need=$md5, found=$file_md5");
}

sub ExecConfFile {
	my @InfoLines;

	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "patch config file is not found.")
		if (! -r $M2K_PATCH_CONF);

	my @InfoLines;

	open (IN, $M2K_PATCH_CONF) || _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "cannot open patch config file: $!");
	@InfoLines = <IN>;
	close(IN);
	
	##################################################################
	##################################################################
	# specified the patch name
	#    PATCH_NAME: <patch-name>
	foreach my $line (@InfoLines) {
		next if($line !~ /^PATCH_NAME:\s+(\S+)\s*$/);

      _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "inconsistent path-name: info=$PATCH_NAME, script=$1")
         if(defined($PATCH_NAME) && $PATCH_NAME ne $1);
         
		$PATCH_NAME = $1;
	}
	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "unknown patch name.")
		if(!defined($PATCH_NAME));
		
	_DEBUG("PATCH-NAME: $PATCH_NAME\n");

	# check patch options
	foreach my $line (@InfoLines) {
		if($line =~ /^ROOT_PERM:\s+([0|1])\s*$/) {
			_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "inconsistent root-perm option: info=$PATCH_ROOT_PERM, script=$1")
   			if($PATCH_ROOT_PERM != int($1));
		}
		if(defined($g_Opt{'a'}) && $line =~ /^POST_INST_MODE:\s+([0|1])\s*$/) {
			_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "inconsistent post-inst-mode option: info=$PATCH_POST_INST_MODE, script=$1")
   			if($PATCH_POST_INST_MODE != int($1));
		}
	}

	# check file md5
	foreach my $line (@InfoLines) {
		next if($line !~ /^FILE_MD5:\s+"([^"]+)"\s+([0-9a-z]+)\s*$/);
		
		_MESG("check MD5 of $1 ($2)...\n");
		Cmd_CheckFileMD5($1, $2);
	}
	
	# check storage space
	#    STORAGE_SPACE: <path> <size>
	foreach my $line (@InfoLines) {
		next if($line !~ /^STORAGE_SPACE:\s+"([^"]+)"\s+(\d+)\s*$/);
		
		_MESG("check storage space of $1 ($2 kbytes)...\n");

		Cmd_CheckStroageSpace($1, int($2));
	}
	
	# check daemons
	#    EFFECT_DAEMON: <daemon-name>
	foreach my $line (@InfoLines)	{
		next if($line !~ /^EFFECT_DAEMON:\s+(\S+)\s*$/);
		
		_MESG("check daemon $1...\n");
		Cmd_CheckDaemon($1);
	}

	# patch dependencies
	#    DEPEND_PATCH: <patch-name>
	foreach my $line (@InfoLines) {
		next if($line !~ /^DEPEND_PATCH:\s+(\S+)\s*(.*)?$/);

		_MESG("check patch dependency: $1...\n");
		Cmd_CheckPatchDep($1, $2);
	}
	
	# file dependencies
	#    DEPEND_FILE: "<file>" <version>
	foreach my $line (@InfoLines) {
		next if($line !~ /^DEPEND_FILE:\s+"([^"]+)"\s+(\S+)\s*$/);

		_MESG("check file dependency: $1 (ver=$2)...\n");
		Cmd_CheckFileDep($1, $2);
	}

	# do backup
	#    BACKUP_FILE: "<file>"
	if (defined($g_Opt{'n'})) {
	   _MESG("ignore backup files (-n specified).");
	}
	else {
      my @BackupFileList;
      foreach my $line (@InfoLines) {
         next if($line !~ /^BACKUP_FILE:\s+"([^"]+)"\s*$/);
         push(@BackupFileList, $1);	
      }
      _MESG("do backup for $#BackupFileList files...\n");
      BackupFile(@BackupFileList);
   }
	
	##################################################################
	##################################################################
	
	_MESG("install patch files...\n");

	my $bIgnoreNext = 0;
	foreach my $line (@InfoLines) {
		chomp $line;
		
		# ignore comment and empty lines
		next if($line =~ /^\s*(#|$)/);

		# ignore control lines
		next if($line =~ /^(PATCH_NAME|EFFECT_DAEMON|DEPEND_PATCH|DEPEND_FILE|STORAGE_SPACE|FILE_MD5|BACKUP_FILE|ROOT_PERM|POST_INST_MODE):/);
		
		if($bIgnoreNext) {
			_DEBUG("ignored-by-IF: $line\n");
			$bIgnoreNext = 0;
			next;
		}

		_DEBUG("PARSING-LINE: $line\n");
		
		if($line =~ /^CREATE_DIR:\s+(\d)\s+(\d{4})\s+"([^"]+)"\s*$/) {
			# CREATE_DIR: <ignore-error-flag> <octal-perm> "<dir-path>"
			Cmd_CreateDir($1, $2, $3);
		}
		elsif($line =~ /^CREATE_DIR_H:\s+(\d)\s+(\d{4})\s+"([^"]+)"\s+(\S+)\s+(\S+)\s*$/) {
			# CREATE_DIR_H: <ignore-error-flag> <octal-perm> "<dir-path>" <owner> <group>
			Cmd_CreateDirHier($1, $2, $3, $4, $5);
		}
		elsif($line =~ /^COPY_FILE:\s+(\d)\s+(\d)\s+"([^"]+)"\s+"([^"]+)"\s*$/) {
			# COPY_FILE: <ignore-error-flag> <keep-property-flag> "<src-path>" "<dest-path>"
			Cmd_CopyFile($1, $2, $3, $4);
		}
		elsif($line =~ /^MOVE_FILE:\s+(\d)\s+(\d)\s+"([^"]+)"\s+"([^"]+)"\s*$/) {
			# MOVE_FILE: <ignore-error-flag> <keep-property-flag> "<src-path>" "<dest-path>"
			Cmd_MoveFile($1, $2, $3, $4);
		}
		elsif($line =~ /^REMOVE_FILE:\s+(\d)\s+"([^"]+)"\s*$/) {
			# REMOVE_FILE: <ignore-error-flag> "<path>"
			Cmd_RemoveFile($1, $2);
		}
		elsif($line =~ /^SET_PERM:\s+(\d)\s+"([^"]+)"\s+(\S+)\s+(\S+)\s+(\d{4})\s*$/) {
			# SET_PERM: <ignore-error-flag> "<path>" <owner> <group> <perm>
			Cmd_SetPerm($1, $2, $3, $4, $5);
		}
		elsif($line =~ /^CREATE_LINK:\s+(\d)\s+(\d)\s+"([^"]+)"\s+"([^"]+)"\s*$/) {
			# CREATE_LINK: <ignore-error-flag> <keep-property-flag> "<path-from>" "<path-to>"
			Cmd_CreateLink($1, $2, $3, $4);
		}
		elsif($line =~ /^EXEC_CMD:\s+(\d)\s+(.*)\s*$/) {
			# EXEC_CMD: <ignore-error-flag> <cmd>
			Cmd_ExeCmd($1, $2);
		}
		elsif($line =~ /^EXEC_CMD_I:\s+(\d)\s+(.*)\s*$/) {
			# EXEC_CMD_I: <ignore-error-flag> <cmd>
			Cmd_ExeCmdByInstaller($1, $2);
		}
		elsif($line =~ /^MESSAGE: (.*)$/) {
			# MESSAGE: <message>
			print "[INFO] ", $1, "\n";
		}
		elsif($line =~ /^IF:\s+(.*)$/) {
			# IF: <if-eval-content>
			$bIgnoreNext = !Cmd_EvalIF($1);
		}
		elsif($line =~ /^STOP:\s+(.*)$/) {
			_FINISH(FIN_S_ERR_USER_ABORT, 0, $1)
		}
		elsif($line =~ /^UMASK:\s+(\d{4})\s*$/) {
			# UMASK: <4-digit-octal>
			umask oct($1);
		}
		else {
			_WARN("unknown line: $line (skipped)\n");
			next;
		}
	}
}

sub BackupFile {
	my @BackupFileList = @_;
	my $ret;
	
	# check backup directory
	if(! -e $M2K_BACKUP_BASEDIR) {
		$ret = s_RunCmd("$MKDIR_CMD -m 0750 -p \"$M2K_BACKUP_BASEDIR\"");
	
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot create backup directory: $M2K_BACKUP_BASEDIR")
			if($ret != 0);

		$ret = s_RunCmd("$CHOWN_CMD $OWNER:$GROUP \"$M2K_BACKUP_BASEDIR\"");
	
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot change owner/group of backup directory: $M2K_BACKUP_BASEDIR")
			if($ret != 0);
	}

	_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "$M2K_BACKUP_BASEDIR is not a directory.")
		if(! -d $M2K_BACKUP_BASEDIR);
	
	# check backup space, extra 5120kb for safty
	my $free_space_needed = 5120 * 1024;
	map {$free_space_needed += (-r $_ ? (-s $_) : 0)} @BackupFileList;
	$free_space_needed = $free_space_needed / 1024;

	my ($mount_point, $free_space) = s_GetStorageFreeSpace(FIN_S_ERR_BACKUP, $M2K_BACKUP_BASEDIR);
	
	_DEBUG("backup space check: need=$free_space_needed, now=$free_space, mount_point=$mount_point\n");
	_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "not enough storage space to make backup, need=$free_space_needed, now=$free_space")
		if($free_space < $free_space_needed);
	
	# decide path of backup file	
	my ($BackupFile, $BackupList, $BackupName) = (undef, undef, undef);
	for(my $n=0; $n<128; $n++) {
		$BackupName = sprintf("m2kpatch_backup_$PATCH_NAME%s.tar", ($n ? "_$n" : ""));
		$BackupList = "$M2K_TMP_DIR/$BackupName.list";
		$BackupFile = "$M2K_BACKUP_BASEDIR/$BackupName";
		last if(! -e $BackupFile && ! -e "$BackupFile.gz");
	}
	_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot decide path of backup file.")
		if(!defined($BackupFile));

	unlink($BackupList) if (-e $BackupList);

	_DEBUG("backup file: $BackupFile\n");
	_DEBUG("backup list: $BackupList\n");

	# output backup file list
	if(!open(FP, "> $BackupList")) {
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot open backup file list: $BackupList: $!");
	}

	my $file_cnt = 0;
	foreach my $file (@BackupFileList) {
		if(! -r $file) {
			_DEBUG("file to backup is not found: $file\n");
			next;
		}
		_DEBUG("output to backup list: $file\n");

		print FP $file, "\n";
		$file_cnt++;
	}

	close(FP);
	chown($OWNER_UID, $GROUP_GID, $BackupList);

	if($file_cnt == 0) {
		_MESG("no file need to backup.\n");
	}
	else {
		# do backup
		my $cmd = "$TAR_CMD -c -T $BackupList -f $BackupFile";

		if (!$DEBUG_MODE) {
			$ret = s_RunCmd("$cmd > /dev/null 2>&1");
		} else {
			$ret = s_RunCmd($cmd);
		}
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot make backup of files in $BackupList, ret=$ret")
			if($ret != 0);

		$ret = chown($OWNER_UID, $GROUP_GID, $BackupFile);

		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot change ownership of $BackupFile, ret=$ret")
			if($ret == 0);

		$ret = s_RunCmd("($GZIP_CMD -f $BackupFile) > /dev/null 2>&1");
		
		_FINISH(FIN_R_ERROR, FIN_S_ERR_BACKUP, "cannot compress backup file: $BackupFile")
			if($ret != 0);

		_MESG("backup: $BackupFile.gz ($file_cnt files), ". (-s "$BackupFile.gz") ." bytes\n");
	}

	unlink($BackupList) if (-e $BackupList);
}

sub UpdateSysPatchInfo {
	my $ret;
	my $ret_code;
	my $patch_profile = "$PATCH_BASEDIR/$M2K_PATCH_PROFILE";

	# old style: update m2kpatch.info
	if(open(FP, ">> $M2K_PATCH_INFO")) {
		printf FP "$PATCH_NAME\t%04d/%02d/%02d %02d:%02d:%02d\n", $_YEAR+1900, $_MON+1, $_MDAY, $_HOUR, $_MIN, $_SEC;
		close(FP);

		$ret = s_RunCmd("$CHOWN_CMD $DEF_OWNER:$DEF_GROUP $M2K_PATCH_INFO");
		_WARN("cannot change owner of $M2K_PATCH_INFO, ret=$ret\n")
			if($ret != 0);
	}
	else {
		_WARN("cannot open patch info file: $M2K_PATCH_INFO: $!\n");
	}
	s_RunCmd("sed -i -e '/if(stForm.opennw.checked)/i \\\tChkCheck();' $M2K_WEBMAIL_HOME/httpd/data/index*.html");
	s_RunCmd("sed -i -e '/libm2k.js/i \\\t\t<script src=\"/j60/libm2k_n.js\" type=\"text/javascript\"></script>' $M2K_WEBMAIL_HOME/httpd/data/index*.html");
	s_RunCmd("chown webmail /webmail/httpd/cgi-bin/dc_ofile");
	s_RunCmd("chgrp webmail /webmail/httpd/cgi-bin/dc_ofile");
	s_RunCmd("chmod 750 /webmail/httpd/cgi-bin/dc_ofile");
	# new style: update patch database
	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_POST, "cannot execute mail2000 patch info tool.")
		if(! -x $M2K_PATCH_TOOL);
	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_POST, "cannot find patch profile: $patch_profile")
		if(! -e $patch_profile);

   # update patch info: patch information to updated patch list (only in manual mode)
   if(!defined($g_Opt{'a'})) {
		if ($RUNNER_UID == 0) {
   		$ret = `su $OWNER -c '$M2K_PATCH_TOOL -a \"$patch_profile\"'`;
		} else {
   		$ret = `$M2K_PATCH_TOOL -a \"$patch_profile\"`;
		}

   	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_POST, "mail2000 patch info tool executed failed (step-1)")
   		if(0 != WEXITSTATUS($?));
   }

   # update patch info: last updated patch information
	if ($RUNNER_UID == 0) {
		$ret = `su $OWNER -c '$M2K_PATCH_TOOL -u \"$patch_profile\"'`;
	} else {
		$ret = `$M2K_PATCH_TOOL -u \"$patch_profile\"`;
	}

	_FINISH(FIN_R_ERROR, FIN_S_ERR_PATCH_POST, "mail2000 patch info tool executed failed (step-2)")
		if(0 != WEXITSTATUS($?));
}

sub ReloadSharedMemory {
   if ($RUNNER_UID == 0) {
      `su $OWNER -c '$M2K_SHM_TOOL'`;
   } else {
      `$M2K_SHM_TOOL`;
   }
}

sub CheckSystem {
	my @stat;

	$RUNNER = "root" if($PATCH_ROOT_PERM);

	_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "system user ($RUNNER) does not exist.")
		if(0 > ($RUNNER_UID = getpwnam($RUNNER)));

	_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "system user ($OWNER) does not exist.")
		if(0 > ($OWNER_UID = getpwnam($OWNER)));
		
	_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "system group ($GROUP) does not exist.")
		if($GROUP ne "" && 0 > ($GROUP_GID = getgrnam($GROUP)));
	
	_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "please run as $RUNNER.")
		if ($< != $RUNNER_UID);

	if (-d $M2K_WEBMAIL_HOME) {
		@stat = stat($M2K_WEBMAIL_HOME);
		_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "the ownership of diretcory ($M2K_WEBMAIL_HOME) should be ($OWNER:$GROUP).")
			if($stat[4] != $OWNER_UID || $stat[5] != $GROUP_GID);
	} else {
		_FINISH(FIN_R_ERROR, FIN_S_ERR_SYSCHK, "the webmail diretcory ($M2K_WEBMAIL_HOME) does not exist.")
	}
}

sub ParsePatchInfo {
   _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "$M2K_PATCH_PROFILE is not found.")
      if(! -r "$M2K_PATCH_PROFILE");

   open(FP, $M2K_PATCH_PROFILE) ||
      _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "cannot open $M2K_PATCH_PROFILE");
      
   while(<FP>) {
      next if(!/^\s*(IsRootInstall|Name|IsPostInstall|Description_en)\s*=\s*(.*?)\s*$/i);
      my ($key, $value) = ($1, $2);
      
      if($key =~ /^IsRootInstall$/i) {
         $PATCH_ROOT_PERM = (int($value) == 0 ? 0 : 1);
      } elsif($key =~ /^IsPostInstall$/i) {
         $PATCH_POST_INST_MODE = (int($value) == 0 ? 0 : 1);
      } elsif($key =~ /^Name$/i) {
         $PATCH_NAME = $value;
      } elsif($key =~ /^Description_en$/i) {
         $PATCH_DESC = $value;
      }
   }
	close(FP);
	
	_DEBUG("$M2K_PATCH_PROFILE: PATCH_ROOT_PERM = $PATCH_ROOT_PERM\n");
	_DEBUG("$M2K_PATCH_PROFILE: PATCH_POST_INST_MODE = $PATCH_POST_INST_MODE\n");
	_DEBUG("$M2K_PATCH_PROFILE: PATCH_NAME = $PATCH_NAME\n");
	_DEBUG("$M2K_PATCH_PROFILE: PATCH_DESC = $PATCH_DESC\n");
}

sub usage {
   print(<<EOF
Usage: $0 <options>
   info:
      -a <result-file>: auto-update mode, output result to result file

   switches:
      -n: do not backup anything (carefully!)
      -P: execute the patch in post-install mode
      -v: verbose mode (output debug messages)
      -q: quiet mode (do not output install messages)
      -h: this help

EOF
);   
   exit;
}

sub main {
   my $post_script_file = undef;

	print STDOUT $VERSION;

	getopts('a:Phnvq', \%g_Opt);
	
	usage() if(defined($g_Opt{'h'}));

	$DEBUG_MODE=1 if(defined($g_Opt{'v'}));
	$QUIET_MODE=1 if(defined($g_Opt{'q'}));

	if(defined($g_Opt{'a'})) {
		$PATCH_STATUS_FILE = $g_Opt{'a'};
	
		_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "cannot write status file: $PATCH_STATUS_FILE")
			if(-e $PATCH_STATUS_FILE && ! -w $PATCH_STATUS_FILE);
	}

	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "please run installer in the patch directory.")
		if(! -e "m2kpatch_ident.txt");

   # parse patch.info
   ParsePatchInfo();
	
	_DEBUG("PATCH_BASEDIR: $PATCH_BASEDIR\n");
	_DEBUG("PATCH_POST_INST_MODE: $PATCH_POST_INST_MODE\n");
	_DEBUG("PATCH_ROOT_PERM: $PATCH_ROOT_PERM\n");
	_DEBUG("PATCH_STATUS_FILE: [$PATCH_STATUS_FILE]\n");
	
	if(!defined($g_Opt{'a'}) && $PATCH_POST_INST_MODE) {
	   _DEBUG("not in auto-update mode, post-install mode has been disabled.\n");
	   $PATCH_POST_INST_MODE = 0;
	}
	
	_MESG("##\n");
	_MESG("## $PATCH_DESC\n");
	_MESG("##\n");
	_MESG("\n");

	if($PATCH_POST_INST_MODE && !defined($g_Opt{'P'})) {
	   # the patch supports post-install mode, but is NOT in post-install mode
      $post_script_file = s_GetPostInstScriptPath();
   	_MESG("** init post-script...\n");
   	_DEBUG("post-script: $post_script_file\n");
   	
   	if(-e $post_script_file) {
   	   _DEBUG("post-script file exists, remove it.\n");
   	   unlink($post_script_file);
   	}
   	
   	if(!open(FP, "> $post_script_file")) {
   	   _FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "cannot open post-script file: $post_script_file: $!");
   	}
   	print FP "#!/bin/sh\n";
   	print FP "# post-script for patch $PATCH_NAME\n";
   	print FP "cd $PATCH_BASEDIR\n";
   	print FP "\n";
   	close(FP);
   	
   	chmod 0755, $post_script_file;
   	
   	_FINISH(FIN_R_ERROR, FIN_S_ERR_INTERNAL, "init post-script failed.")
      	if(! -x $post_script_file);
   }

	_MESG("** check system...\n");
	CheckSystem();

   if($PATCH_POST_INST_MODE && defined($g_Opt{'P'})) {	
      # the patch supports post-install mode, and execute in post-install mode
      $post_script_file = s_GetPostInstScriptPath();
      
      if(! -x $post_script_file) {
         _DEBUG("cannot execute post-script: $post_script_file\n");
      }
      else {
         _MESG("** execute post-script...\n");
         s_RunCmd("sh $post_script_file");
      }
   }
   else {
      # the patch does NOT support post-install mode, or
      # the patch supports post-install mode, but is NOT in post-install mode
   	_MESG("** apply patch...\n");
   	ExecConfFile();
   }

	if($PATCH_POST_INST_MODE && !defined($g_Opt{'P'})) {
	   # the patch supports post-install mode, but is NOT in post-install mode
	   _DEBUG("post-script is ready: $post_script_file");
      _FINISH(FIN_R_SUCC, FIN_S_ERR_NONE, "post-script is ready: $post_script_file");
   }
   else {
      # the patch does NOT support post-install mode, or
      # the patch supports post-install mode, and run in post-install mode
      _MESG("** update patch database...\n");
      UpdateSysPatchInfo();

      _MESG("** reload system config...\n");
      ReloadSharedMemory();

      if($PATCH_POST_INST_MODE && defined($g_Opt{'P'})) {
         _MESG("** remove post-install script...\n");
         unlink($post_script_file);
      }

      _FINISH(FIN_R_SUCC, FIN_S_ERR_NONE, "patch installed successfully.");
   }
}

exit main();
